 /* ops.c   --  heightfield operations
 *              Jan 15 1996   John P. Beale
 */

#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <math.h>                       /* math stuff */
#include <string.h>                     /* strcpy() */
#include <ctype.h>                      /* tolower() */
#include <dirent.h>                     /* opendir, readdir */
#ifndef NOTIMEH
#include <time.h>                       /* time for seed */
#endif
#include "hl.h"                         /* most fn, #def'ns */
#include "fftn.h"                       /* fft, M_PI */

extern hfield *hst;            /* stack of heightfields */
extern int hp;                 /* pointer to top of stack: hst[hp] */
extern int stacksize;          /* total size of stack */
extern int histbins;           /* histogram bins */
extern int rand_gauss;         /* if gaussian rand# should be used */
extern int rnd_seed_stale;     /* no random seed yet? */
extern int rnd_seed;           /* random seed to use (if not stale) */
extern time_t time_start;       /* starting time, set by time_tic() */

static double arand, gaussadd, gaussfac; /* Gaussian random parameters */

#define Nrand 4            /* number of rand samples for gaussian */

/* ------- functions defined in this file ------------ */
int time_tic();      /* store current time in time_start global */
int time_toc();      /* print difference between current time and tic_time */
void initgauss();
double gauss();
double gaussn();      /* returns gaussian dist. rand # between 0 and 1 */
int smooth(D frac);
int gen_rand(int xsize, int ysize);
int h_fft(int dir, D s);
int gforge(int size, float h);
int fillarray(int xsize, int ysize, float h);
int yslope(D yfrac,D scale, D slope_exp);
int h_gauss(D xfrac, D yfrac, D radfac, D hscale);  /* add gaussian hill */
int h_ring(D xfrac, D yfrac, D radfac, D width, D hscale);  /* add a ring */
int h_crater(int how_many, D ch_scale, D radius, D dfac); /* add craters */

int h_fourfilt(D center, D Q, char *f_type);  /* filter real & imag. matrix */
void f_filter(PTYPE *real, PTYPE *imag, int xsize, int ysize, 
	      double center, double Q, int f_type);

/* --------------------------------------------------- */

/* time_tic()  save current time for later use by time_toc() */
/*             uses global variable   time_start             */
int time_tic()
{
  time_start = time(NULL);  /* elapsed time in seconds from Jan 1970 */
  return(0);
}

int time_toc()  /* print out elapsed time since call to time_tic() */
{
time_t elapsed;

  elapsed = time(NULL) - time_start;
  printf("%ld seconds.\n",(long)elapsed);
  return(0);
}

/* --------------------------------------------------- */

/* smooth(f) -- smooth matrix by replacing each HF element with average of
	    its neighbors.  f=0: no smoothing. f=1.0 : full smoothing.

   globals: -----------------
   hst[]        HF stack
   hp           HF stack pointer
 */

int smooth(D frac)
{
 int x,y;
 int xsize,ysize;
 double tmp, orig;
 hfield h0, h1;
 int wrap;

  if ((hst[hp].a) == (PTYPE *)NULL) {
     printf("Stack empty.\n");
     return(1);
  }
  
  h0 = hst[hp];
  if (h0.c) printf("Note: smoothing real part only.\n");
  xsize = h0.xsize;
  ysize = h0.ysize;
  wrap = h_tilable(h0,0);           /* if image is tilable or not */
  
  if (h0.c) {
    if (h_newc(&h1,xsize,ysize,h0.name)) return(1);
  } else {
    if (h_newr(&h1,xsize,ysize,h0.name)) return(1);
  }

  for(y=0;y<ysize;y++) {
    for (x=0;x<xsize;x++) {
      orig = El(h0.a,x,y);
      if (wrap) tmp = Elmod(h0.a,x-1,y) + Elmod(h0.a,x,y-1)+
		      Elmod(h0.a,x+1,y) + Elmod(h0.a,x,y+1);
      else      tmp = Elclip(h0.a,x-1,y) + Elclip(h0.a,x,y-1)+
		      Elclip(h0.a,x+1,y) + Elclip(h0.a,x,y+1);

      tmp /= 4;         /* compute average of neighbors */
      tmp = orig + frac*(tmp-orig);
      El(h1.a,x,y) = tmp;
    }
  }
  free(h0.a);    /* free old memory */
  hst[hp] = h1;  /* put new HF on stack */
  h_minmax();    /* update min, max */
  return(0);
}  /* smooth() */

int gen_rand(int xsize, int ysize)
{
 int x,y;
 hfield h1;
 PTYPE *hf;
 D hmin,hmax,tmp;
 
   if (h_newr(&h1,xsize,ysize,"rand")) return(1);
   hf = h1.a;

   initgauss(); 
   hmin = FLT_MAX;
   hmax = FLT_MIN;
   for (y=0;y<ysize;y++) {
     for (x=0;x<xsize;x++) {
      if (rand_gauss) tmp = gaussn();
       else tmp = ran1();
      if (tmp > hmax) hmax = tmp;
      if (tmp < hmin) hmin = tmp;
      El(hf,x,y) = tmp;
     }
   }
   h1.min = hmin;
   h1.max = hmax;

   INCSTACK;                              /* move up one on stack */
   if (hst[hp].a != (PTYPE *)NULL) free(hst[hp].a);
   hst[hp] = h1;
   return(0);
} /*-------------------  gen_rand() ---------- */

/* h_fft(dir, scal)  -- do forward or inverse fourier transform
 *                 dir = +1  forward FFT
 *                       -1  inverse FFT
 *                 scal =  1.0  no renormalization
 *                 scal = -1.0  renorm by mx. dimension
 *                 scal < -1.0  renorm by sqrt(mx dim)
 */

int h_fft(int dir, D scaling)
{
int dim[2];           /* fft dimensions */
int xsize,ysize;
hfield hf;  
PTYPE *hf0, *hf1;
 
 if (s_empty()) return(1);     /* check if stack is empty */
 hf = hst[hp];
 if (!hf.c) {
  printf("fft: error: complex matrix is required.\n");
  return(1);
 }
 xsize = hf.xsize;
 ysize = hf.ysize;
 hf0 = hst[hp].a;
 hf1 = &(hst[hp].a[xsize*ysize]);  /* starting point of imag. matrix */
 dim[0] = xsize;
 dim[1] = ysize;

 /* #dims dim_array, *real, *imag, fwd/rev, scaling */
 /* if scaling = -1 norm by dimension, scaling < -1 norm by sqrt(dim) */

 fftnf(2, dim, hf0, hf1, dir, scaling); 
 h_minmax();
 return(0);
} /* h_fft() */

int gforge(int size, float h)
{
 char s[20] = {'g','f'};
 
 if (size<3) {printf("Gforge error: minimum array size is 3.\n"); return(1);}
 if ((h<0)||(h>4)) {
   printf("Gforge error: dimension must be in range of [0..4]\n");
   return(1);
 }
 printf("filling array..."); 
 fflush(stdout);
 if (fillarray(size,size,3.0-h)) return(1); /* generate complex array */
 printf("taking IFFT...");
 fflush(stdout);
 h_fft(-1,1.0);                     /* take inverse fft */
 printf(".\n");
 c_real();                      /* separate real & imag. parts */
 norm(0,1);                     /* normalize to 0..1 */
 sprintf(&s[2],"%1.2f",h);
 strcpy(hst[hp].name,s);         /* give a reasonable name */
 return(0);
}

	/* fill array with 1/f gaussian noise */
int fillarray(int xsize, int ysize, float h)
{
 hfield h1;
 int x,y, k, i0, j0, rank, nx, ny, xcent, ycent, rankmax;
 double rad, phase, rcos, rsin, scale;
 PTYPE *real, *imag;

 if (h_newc(&h1,xsize,ysize,"1/fnoise")) return(1);

 nx = xsize; ny = ysize;
 real = h1.a;
 imag = &(h1.a[xsize*ysize]);

 /* printf("filling %d x %d HF\n",xsize,ysize); */

 xcent = (int)(xsize / 2.0 - 0.5);  /* center dimensions of array */
 ycent = (int)(ysize / 2.0 - 0.5);
 rankmax = MIN(xcent,ycent);

    /* fill in mx. in order of radius, so we can generate higher resolutions
       with the same overall aspect (if seed is held fixed) */
 
			 /* initialize gaussian/random # with time */
    initgauss();
    scale = 1.0;

    for (rank = 0; rank <= rankmax; rank++) {
    /* fill quadrants 2 and 4  */
      for (k=0;k<=rank;k++) {
       x = k; y = rank;
       phase = 2 * M_PI * ((ran1() * 0x7FFF) / arand);
       if ((x == 0) && (y == 0)) rad = 0; 
	 else rad = pow((double) (x*x + y*y), -(h+1) / 2) * gauss();
       rcos = rad * cos(phase)*scale; rsin = rad * sin(phase)*scale;
       El(real, x, y) = rcos; 
       El(imag, x, y) = rsin;
       if (!((x == 0) && (y == 0))) { 
	i0 = nx-x-1; j0 = ny-y-1;
	El(real, i0,j0) = rcos;
	El(imag, i0,j0) = rsin;
       }

       x = rank; y = k;
       phase = 2 * M_PI * ((ran1() * 0x7FFF) / arand);
       if ((x == 0) && (y == 0)) rad = 0; 
	 else rad = pow((double) (x*x + y*y), -(h+1) / 2) * gauss();
       rcos = rad * cos(phase)*scale; rsin = rad * sin(phase)*scale;
       El(real, x, y) = rcos;
       El(imag, x, y) = rsin;
       if (!((x == 0) && (y == 0))) { 
	i0 = nx-x-1; j0 = ny-y-1;
	El(real, i0,j0) = rcos;
	El(imag, i0,j0) = rsin;
       }
      } /* end for k */
    
    /* now handle quadrants 1 and 3 */
      for (k=0;k<=rank;k++) {
       x = k; y = rank;
       phase = 2 * M_PI * ((ran1() * 0x7FFF) / arand);
       if ((x == 0) && (y == 0)) rad = 0; 
	 else rad = pow((double) (x*x + y*y), -(h+1) / 2) * gauss();
       rcos = rad * cos(phase)*scale; 
       rsin = rad * sin(phase)*scale;
       El(real, x, ny-y-1) = rcos; 
       El(imag, x, ny-y-1) = rsin;
       El(real, nx-x-1, y) = rcos;
       El(imag, nx-x-1, y) = rsin;
       
       x = rank; y = k;
       phase = 2 * M_PI * ((ran1() * 0x7FFF) / arand);
       if ((x == 0) && (y == 0)) rad = 0; 
	 else rad = pow((double) (x*x + y*y), -(h+1) / 2) * gauss();
       rcos = rad * cos(phase)*scale; rsin = rad * sin(phase)*scale;
       El(real, x, ny-y-1) = rcos; 
       El(imag, x, ny-y-1) = rsin;
       El(real, nx-x-1, y) = rcos;
       El(imag, nx-x-1, y) = rsin;
      
      } /* end for k */
    } /* end for rank */
    
    El(imag, nx / 2, 0) = 0;
    El(imag, 0, ny / 2) = 0;
    El(imag, nx / 2, ny / 2) = 0;
 
  INCSTACK;
  if (hst[hp].a != (PTYPE *)NULL) free(hst[hp].a);
  hst[hp] = h1;
  h_minmax();  /* find limits of real array */
  return(0);
} /* end fillarray() */

/*  INITGAUSS  --  Initialise random number generators.  As given in
		   Peitgen & Saupe, page 77. */

void initgauss()
{
    /* Range of random generator */
    arand = pow(2.0, 15.0) - 1.0;
    gaussadd = sqrt(3.0 * Nrand);
    gaussfac = 2 * gaussadd / (Nrand * arand);

    if (rnd_seed_stale)  /* use time of day as seed, if stale */
      rnd_seed = (int) (time(NULL) ^ 0xF37C)%1000000;
    else		/* otherwise, just increment seed by one */
      rnd_seed = (rnd_seed+1)%1000000;

    seed_ran1((int)rnd_seed);        /* seed the generator */
    rnd_seed_stale = FALSE;
}

/*  GAUSS  --  Return a Gaussian random number.  As given in Peitgen
	       & Saupe, page 77. */
double gauss()
{
 int i;
 D sum = 0.0;

 for (i=1; i<=Nrand; i++) sum += (ran1() * 0x7FFF);
 return(gaussfac * sum - gaussadd);
}

double gaussn()   /* returns gaussian dist. rand # between 0 and 1 */
{
 int i;
 D sum = 0.0;

 for (i=0; i<Nrand; i++) sum += ran1();
 return(sum/Nrand);
}

int yslope(D yfrac,D y_sfac, D slope_exp)
{
 int xsize,ysize;
 int backedge = TRUE;   /* if slope is max at back edge of hf */
 int i,j;
 D yf, z;
 D yscale;      /* num to multiply yf by so range = 0..1 */
 PTYPE *hf;
 PTYPE hmin, hmax;

 if (s_empty()) return(1);     /* check if stack is empty */
 if ((yfrac < 0)||(yfrac > 1)) {
   printf("yslope: frac must be within [0..1].\n");
   return(1);
 }
 if (slope_exp<0) {
   printf("yslope: exponent must be positive.\n");
   return(1);
 }
 xsize = hst[hp].xsize;
 ysize = hst[hp].ysize;
 hf = hst[hp].a;
 hmin = FLT_MAX;
 hmax = FLT_MIN;

 yfrac = 1-yfrac;
 if (yfrac == 0) yscale = 1;
   else yscale = 1/yfrac;

 for (j=0;j<ysize;j++) {
   for (i=0;i<xsize;i++) {
    yf = ((double)j)/ysize;   /* fractional distance along y coord */
    if (backedge) yf = 1-yf;
    yf = (yf - yfrac) * yscale;
    if (yf < 0) yf = 0;
    z = (y_sfac * pow(yf,slope_exp)) + El(hf,i,j);
    El(hf,i,j) = (PTYPE)z;
    if (z > hmax) hmax = z;
    if (z < hmin) hmin = z;
   } /* end for i */
  } /* end for j */
  hst[hp].min = hmin;
  hst[hp].max = hmax;

 return(0);
}  /* yslope */

int h_gauss(D xfrac, D yfrac, D radfac, D hscale)  /* add gaussian hill */
{
 long int xsize,ysize;
 long int xmin, xmax, ymin, ymax;
 long int x,y;
 D xa,ya,tmp,rad,xcent,ycent;
 PTYPE *hf;
 PTYPE hmin, hmax;
 int tile;

 if (s_empty()) return(1);     /* check if stack is empty */
 if ((yfrac < 0)||(yfrac > 1)||(xfrac<0)||(yfrac>1)) {
   printf("gauss: xpox, ypos must be within [0..1].\n");
   return(1);
 }
 xsize = hst[hp].xsize;
 ysize = hst[hp].ysize;
 hf = hst[hp].a;
 hmin = FLT_MAX;
 hmax = FLT_MIN;
 xcent = xsize/2.0;  ycent = ysize/2.0;
 tile = is_tilable(0);
 if (tile) {
   xmin = xcent-(GAUFAC*radfac*xsize); 
   xmax = xcent+(GAUFAC*radfac*xsize);
   ymin = ycent-(GAUFAC*radfac*ysize); 
   ymax = ycent+(GAUFAC*radfac*ysize);;
 } else {
   xmin = 0; xmax = xsize;
   ymin = 0; ymax = ysize;
 }
 radfac = radfac * radfac;
 for (y=ymin;y<ymax;y++) {
   ya = (((D)y+0.5)/ysize)-yfrac;
   ya = ya*ya;
   for (x=xmin;x<xmax;x++) {
/*    xx = (x>=0)?(x%xsize):(((x%xsize)+xsize))%xsize;
    if ( ((x+10*xsize)%xsize==0) && ((y+10*ysize)%ysize==0) ) 
      printf("%ldx%ld ",xx,yy); */
    if (tile) tmp = Elmod(hf,x,y); else tmp = El(hf,x,y); 
/*    if (tile) tmp = El(hf,xx,yy); else tmp = El(hf,x,y); */
    xa = (((D)x+0.5)/xsize)-xfrac;
    xa = xa * xa;
    rad = xa+ya;
    tmp += hscale * exp(-rad/radfac);
    if (tile) Elmod(hf,x,y) = tmp; else El(hf,x,y) = tmp; 
    if (tmp > hmax) hmax = tmp;
    if (tmp < hmin) hmin = tmp;
   } /* end for x */
  } /* end for y */
  if (tile) h_minmax();
  else {
    hst[hp].min = hmin;
    hst[hp].max = hmax;
  }

 return(0);
}  /* h_gauss()  */

int h_ring(D xfrac, D yfrac, D radfac, D width, D hscale)  /* add a ring */
{
 long int xsize,ysize;
 long int xmin, xmax, ymin, ymax;
 long int x,y;
 D xa,ya,tmp,rad,xcent,ycent;
 PTYPE *hf;
 PTYPE hmin, hmax;
 int tile;

 if (s_empty()) return(1);     /* check if stack is empty */
 if ((yfrac < 0)||(yfrac > 1)||(xfrac<0)||(yfrac>1)) {
   printf("gauss: xpox, ypos must be within [0..1].\n");
   return(1);
 }
 xsize = hst[hp].xsize;
 ysize = hst[hp].ysize;
 hf = hst[hp].a;
 hmin = FLT_MAX;
 hmax = FLT_MIN;
 xcent = xsize/2.0;  ycent = ysize/2.0;
 tile = is_tilable(0);
 if (tile) {
   xmin = xcent-((GAUFAC*width+radfac)*xsize); 
   xmax = xcent+((GAUFAC*width+radfac)*xsize);
   ymin = ycent-((GAUFAC*width+radfac)*ysize); 
   ymax = ycent+((GAUFAC*width+radfac)*ysize);;
 } else {
   xmin = 0; xmax = xsize;
   ymin = 0; ymax = ysize;
 }
 width = width * width;  /* gaussian sigma, squared */
 for (y=ymin;y<ymax;y++) {   /* --- main loop over (x,y) ---- */
   ya = (((D)y+0.5)/ysize)-yfrac;
   ya *= ya;
   for (x=xmin;x<xmax;x++) {
    if (tile) tmp = Elmod(hf,x,y); else tmp = El(hf,x,y); 
    xa = (((D)x+0.5)/xsize)-xfrac;
    xa *= xa;
    rad = sqrt(xa + ya);
    tmp += hscale * exp(-pow((rad-radfac)/width,2));  /* gaussian */
    if (tile) Elmod(hf,x,y) = tmp; else El(hf,x,y) = tmp; 
    if (tmp > hmax) hmax = tmp;
    if (tmp < hmin) hmin = tmp;
   } /* end for x */
  } /* end for y */
  if (tile) h_minmax();
  else {
    hst[hp].min = hmin;
    hst[hp].max = hmax;
  }
 return(0);
} /* h_ring() */

/* -  h_crater() ---- add craters to a heightfield ----------- */

int h_crater(int how_many, D ch_scale, D radius, D dfac)
{ PTYPE *real, *imag;
  hfield h0,h1;           /* real and (temporary) imaginary HFs */
  int xsize, ysize;
  int wrap;

  if (s_empty()) return(1);     /* check if stack is empty */
  h0 = hst[hp];
  xsize = h0.xsize;
  ysize = h0.ysize;
  real = h0.a;

  if (h_newr(&h1,xsize,ysize,"i_temp")) return(1);
  imag = h1.a;
  wrap = h_tilable(h0,0);
  initgauss();    /* seed or re-seed random # generator */

  distribute_craters(real, imag, xsize, ysize, (U) how_many, 
			  wrap, ch_scale, radius, dfac);

  free(imag);  /* unload imaginary (temp) array */

  h_minmax();           /* update min, max values */
  return(0);
} /* h_crater() */


/* f_filter()  -- filter array of noise  with peak or notch filter */
/*                a frequency-domain filter: acts on real & imag arrays */
/*                f_type: 1 bpass, -1 breject, 2 lopass, -2 hipass */

void f_filter(PTYPE *real, PTYPE *imag, int xsize, int ysize, 
	      double center, double Q, int f_type)
{
 int i, j, i0, j0;
 double rad, fac, p, sfac;

    if (center == 0.0) center = -0.00001;  /* avoid a singularity */

		     /* sfac is radius scaling factor */
    sfac = 1.0/sqrt((double)(xsize*xsize/4 + ysize*ysize/4));  
    for (i = 0; i <= xsize / 2; i++) {          /* do quadrants 2 and 4 */
	for (j = 0; j <= ysize / 2; j++) {
	    if (i != 0 || j != 0) {
		rad = sqrt((double) (i * i + j * j)) * sfac;
	    } else {
		rad = 0;
	    }
	    p = 1.0 / pow(Q * center, 2);
	    if (abs(f_type)==1)
	      fac = p / (p + pow((1.0-rad/center),2)) ; /* bandpass/rej. */
	    else
	      fac = 1.0 / (1.0 + pow((rad/center),Q) ); /* lo/hi-pass */
	    if (f_type < 0) fac = (1.0 - fac);  /* invert filter */
	    El(real, i, j) *= fac;
	    El(imag, i, j) *= fac;
	    i0 = (i == 0) ? 0 : xsize - i;
	    j0 = (j == 0) ? 0 : ysize - j;
	    El(real, i0, j0) *= fac;
	    El(imag, i0, j0) *= fac;
	}
    }
    El(imag, xsize / 2, 0) = 0;
    El(imag, 0, ysize / 2) = 0;
    El(imag, xsize / 2, ysize / 2) = 0;
    for (i = 1; i <= xsize / 2 - 1; i++) {     /* do quadrants 1 and 3 */
	for (j = 1; j <= ysize / 2 - 1; j++) {
	    rad = sfac * sqrt((double) (i * i + j * j));
	    p = 1.0 / pow(Q * center, 2);
	    if (abs(f_type)==1)
	      fac = p / (p + pow((1.0-rad/center),2)) ; /* bandpass/rej. */
	    else
	      fac = 1.0 / (1.0 + pow((rad/center),Q) ); /* lo/hi-pass */
	    if (f_type < 0) fac = (1.0 - fac);  /* invert filter */
	    El(real, i, ysize - j) *= fac;
	    El(imag, i, ysize - j) *= fac;
	    El(real, xsize - i, j) *= fac;
	    El(imag, xsize - i, j) *= fac;
	}
    }

} /* end f_filter() */

/* h_fourfilt()  -- do filtering on fourier-domain data in real and
 *              imaginary arrays on stack
 *
 *  calls: f_filter(PTYPE *real, PTYPE *imag, int xsize, int ysize, 
 *             double center, double Q, int f_type)
 *
 */

int h_fourfilt(D center, D Q, char *f_type)   /* filter real & imag. matrix */
{
 int type;              /* filter type */
 int xsize,ysize;
 hfield hf;
 PTYPE *imag, *real;
 
 if (s_empty()) return(1);     /* check if stack is empty */
 hf = hst[hp];
 if (!hf.c) {
   printf("fourfilt error: matrix must be complex.\n");
   return(1);
 }
 xsize = hf.xsize;
 ysize = hf.ysize;
 real = hf.a;
 imag = &(hf.a[xsize*ysize]);

 if      (keymatch(f_type,"ffbp",4)) type = BANDPASS;
 else if (keymatch(f_type,"ffbr",4)) type = BANDREJECT;
 else if (keymatch(f_type,"fflp",4)) type = LOWPASS;
 else if (keymatch(f_type,"ffhp",4)) type = HIGHPASS;
 else {
   printf("h_fourfilt: unknown filter type %s\n",f_type);
   return(1);
 }
 
 f_filter(real, imag, xsize, ysize, center, Q, type);  /* do filtering */

 h_minmax();  /* and those of real array */
 return(0);
} /* h_fourfilt() */

/* h_realfilt() -- Fourier filter a real-valued matrix: fft->filter->ifft */

int h_realfilt(D a1, D a2, char *f_type)
{
 int xsize, ysize;
 char fs[6];
 
 if      (keymatch(f_type,"bp",2)) {
   strcpy(fs,"ffbp");
   if (a2==0) a2=4;
 } else if (keymatch(f_type,"br",2)) {
   strcpy(fs,"ffbr");
   if (a1==-1) a1=0.0;
   if (a2==0) a2=50;
 } else if (keymatch(f_type,"lp",2)) {
   strcpy(fs,"fflp");
 } else if (keymatch(f_type,"hp",2)) {
   if (a1==-1) a1 = 0.05;
   strcpy(fs,"ffhp");
 } else {
   printf("h_realfilt: unknown filter type %s\n",f_type);
   return(1);
 }
 if (a1==-1) a1=0.1;
 if (a2==0) a2=1;

 if (s_empty()) return(1);     /* check if stack is empty */
 xsize = hst[hp].xsize;
 ysize = hst[hp].ysize;
 if (c_swap()) return(1);       /* make matrix complex */
 c_swap();
 printf("FFT...");fflush(stdout);
 h_fft(1,1.0);   /* forward FFT, no rescaling */

 printf("filter...");fflush(stdout);
 h_fourfilt(a1,a2,fs);

 printf("IFFT...");fflush(stdout);
 h_fft(-1,-1.0); /* inverse FFT, rescale by mx. size */
 printf("\n");
 c_real();       /* leave just real part on stack */
 return(0);
} /* h_realfilt */
